from typing import Tuple, Union

__all__ = ["VolumeSlicer"]
import numpy as np
from scipy.ndimage import distance_transform_edt
import itertools


class VolumeSlicer:
    """
    Helper class to slice 3d volume into smaller volumes and merge them back
    """

    def __init__(
        self,
        data,
        patch_size: Union[int, Tuple[int, int, int]],
        overlap: Tuple[int, int, int],
        crop: Tuple[int, int, int],
    ):
        """
        :param volume_shape: Shape of the source image (D, H, W, C)
        :param patch_size: Tile size (Scalar or tuple (D, H, W, C)
        :param voxel_step: Step in pixels between voxels (Scalar or tuple (D, H, W))
        :param weight: Fusion algorithm. 'mean' - avergaing
        """
        self.data = data
        self.data_shape = np.array(data.shape)

        if isinstance(patch_size, (tuple, list)):
            if len(patch_size) != 3:
                raise ValueError()
            self.patch_size = np.array(patch_size, dtype=int)
        else:
            self.patch_size = np.array([int(patch_size)] * 3)

        if isinstance(overlap, (tuple, list)):
            if len(overlap) != 3:
                raise ValueError()
            self.overlap = np.array(overlap, dtype=int)
        else:
            self.overlap = np.array([int(overlap)] * 3)

        if isinstance(crop, (tuple, list)):
            if len(crop) != 3:
                raise ValueError()
            self.crop = np.array(crop, dtype=int)
        else:
            self.crop = np.array([int(crop)] * 3)

        assert all(
            [
                p - 2 * o - 2 * c > 0
                for p, o, c in zip(self.patch_size, self.overlap, self.crop)
            ]
        ), "Invalid combination of patch size, overlap and crop size."
        # Calculate the position of each tile
        locations = []
        for i, p, o, c in zip(
            self.data_shape, self.patch_size, self.overlap, self.crop
        ):
            # get starting coords
            coords = (
                np.arange(
                    np.ceil((i + o + c) / np.maximum(p - o - 2 * c, 1)),
                    dtype=np.int16,
                )
                * np.maximum(p - o - 2 * c, 1)
                - o
                - c
            )
            locations.append(coords)
        self.locations = list(itertools.product(*locations))
        self.global_crop_before = np.abs(
            np.min(np.array(self.locations), axis=0)
        )
        self.global_crop_after = (
            np.array(self.data_shape)
            - np.max(np.array(self.locations), axis=0)
            - np.array(self.patch_size)
        )

    def get_fading_map(self):

        fading_map = np.ones(self.patch_size)

        if all([c == 0 for c in self.crop]):
            self.crop = [1, 1, 1]

        # Exclude crop region
        crop_masking = np.zeros_like(fading_map)
        crop_masking[
            self.crop[0] : self.patch_size[0] - self.crop[0],
            self.crop[1] : self.patch_size[1] - self.crop[1],
            self.crop[2] : self.patch_size[2] - self.crop[2],
        ] = 1
        fading_map = fading_map * crop_masking

        fading_map = distance_transform_edt(fading_map).astype(np.float32)

        # Normalize
        self.fading_map = fading_map / fading_map.max()

    def split(self, idx):

        self.patch_start = np.array(self.locations[idx])
        self.patch_end = self.patch_start + np.array(self.patch_size)
        pad_before = np.maximum(-self.patch_start, 0)
        pad_after = np.maximum(self.patch_end - np.array(self.data_shape), 0)
        pad_width = list(zip(pad_before, pad_after))

        slicing = tuple(
            map(slice, np.maximum(self.patch_start, 0), self.patch_end)
        )
        self.tile = self.data[slicing]
        self.tile = np.pad(self.tile, pad_width, mode="reflect")



"""
3D Squeeze and Excitation Modules
*****************************
3D Extensions of the following 2D squeeze and excitation blocks:
    1. `Channel Squeeze and Excitation <https://arxiv.org/abs/1709.01507>`_
    2. `Spatial Squeeze and Excitation <https://arxiv.org/abs/1803.02579>`_
    3. `Channel and Spatial Squeeze and Excitation <https://arxiv.org/abs/1803.02579>`_
New Project & Excite block, designed specifically for 3D inputs
    'quote'
    Coded by -- Anne-Marie Rickmann (https://github.com/arickm)
"""

import torch
from torch import nn as nn
from torch.nn import functional as F


class ChannelSELayer3D(nn.Module):
    """
    3D extension of Squeeze-and-Excitation (SE) block described in:
        *Hu et al., Squeeze-and-Excitation Networks, arXiv:1709.01507*
        *Zhu et al., AnatomyNet, arXiv:arXiv:1808.05238*
    """

    def __init__(self, num_channels, reduction_ratio=2):
        """
        Args:
            num_channels (int): No of input channels
            reduction_ratio (int): By how much should the num_channels should be reduced
        """
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool3d(1)
        num_channels_reduced = num_channels // reduction_ratio
        self.reduction_ratio = reduction_ratio
        self.fc1 = nn.Linear(num_channels, num_channels_reduced, bias=True)
        self.fc2 = nn.Linear(num_channels_reduced, num_channels, bias=True)
        self.relu = nn.ReLU()
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        batch_size, num_channels, D, H, W = x.size()
        # Average along each channel
        squeeze_tensor = self.avg_pool(x)

        # channel excitation
        fc_out_1 = self.relu(
            self.fc1(squeeze_tensor.view(batch_size, num_channels))
        )
        fc_out_2 = self.sigmoid(self.fc2(fc_out_1))

        output_tensor = torch.mul(
            x, fc_out_2.view(batch_size, num_channels, 1, 1, 1)
        )

        return output_tensor


class SpatialSELayer3D(nn.Module):
    """
    3D extension of SE block -- squeezing spatially and exciting channel-wise described in:
        *Roy et al., Concurrent Spatial and Channel Squeeze & Excitation in Fully Convolutional Networks, MICCAI 2018*
    """

    def __init__(self, num_channels):
        """
        Args:
            num_channels (int): No of input channels
        """
        super().__init__()
        self.conv = nn.Conv3d(num_channels, 1, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x, weights=None):
        """
        Args:
            weights (torch.Tensor): weights for few shot learning
            x: X, shape = (batch_size, num_channels, D, H, W)
        Returns:
            (torch.Tensor): output_tensor
        """
        # channel squeeze
        batch_size, channel, D, H, W = x.size()

        if weights:
            weights = weights.view(1, channel, 1, 1)
            out = F.conv2d(x, weights)
        else:
            out = self.conv(x)

        squeeze_tensor = self.sigmoid(out)

        # spatial excitation
        output_tensor = torch.mul(
            x, squeeze_tensor.view(batch_size, 1, D, H, W)
        )

        return output_tensor


class ChannelSpatialSELayer3D(nn.Module):
    """
    3D extension of concurrent spatial and channel squeeze & excitation:
        *Roy et al., Concurrent Spatial and Channel Squeeze & Excitation in Fully Convolutional Networks, arXiv:1803.02579*
    """

    def __init__(self, num_channels, reduction_ratio=2):
        """
        Args:
            num_channels (int): No of input channels
            reduction_ratio (int): By how much should the num_channels should be reduced
        """
        super().__init__()
        self.cSE = ChannelSELayer3D(num_channels, reduction_ratio)
        self.sSE = SpatialSELayer3D(num_channels)

    def forward(self, input_tensor):
        output_tensor = torch.max(
            self.cSE(input_tensor), self.sSE(input_tensor)
        )
        return output_tensor



import numpy as np
import os

# from IPython.display import clear_output
from stardist.models import Config3D, StarDist3D
from stardist import Rays_GoldenSpiral, calculate_extents
from csbdeep.utils import normalize
import glob
from csbdeep.io import load_training_data
from csbdeep.models import Config
from .CARE import CARE
from tensorflow.keras.utils import Sequence
from csbdeep.data import RawData, create_patches
from skimage.measure import label, regionprops
from scipy import ndimage
import matplotlib.pyplot as plt
from pathlib import Path
from tifffile import imread, imwrite
from csbdeep.utils import plot_history
from scipy.ndimage import zoom
from scipy.ndimage import binary_fill_holes


def erode_labels(segmentation, erosion_iterations=2):
    # create empty list where the eroded masks can be saved to
    regions = regionprops(segmentation)
    erode = np.zeros(segmentation.shape)

    def erode_mask(segmentation_labels, label_id, erosion_iterations):

        only_current_label_id = np.where(segmentation_labels == label_id, 1, 0)
        eroded = ndimage.binary_erosion(
            only_current_label_id, iterations=erosion_iterations
        )
        relabeled_eroded = np.where(eroded == 1, label_id, 0)
        return relabeled_eroded

    for i in range(len(regions)):
        label_id = regions[i].label
        erode = erode + erode_mask(segmentation, label_id, erosion_iterations)

    # convert list of numpy arrays to stacked numpy array
    return erode


class SmartSeeds3D:
    def __init__(
        self,
        base_dir,
        unet_model_name,
        star_model_name,
        unet_model_dir,
        star_model_dir,
        npz_filename=None,
        n_patches_per_image=1,
        train_loss="mae",
        raw_dir="raw/",
        real_mask_dir="real_mask/",
        binary_mask_dir="binary_mask/",
        val_raw_dir="val_raw/",
        val_real_mask_dir="val_real_mask/",
        n_channel_in=1,
        n_channel_out=1,
        pattern=".tif",
        downsample_factor=1,
        backbone="resnet",
        load_data_sequence=True,
        train_unet=True,
        train_star=True,
        generate_npz=True,
        validation_split=0.01,
        erosion_iterations=2,
        patch_size=(16, 256, 256),
        grid_x=1,
        grid_y=1,
        annisotropy=(1, 1, 1),
        use_gpu=True,
        val_size: int = None,
        batch_size=4,
        depth=3,
        kern_size=3,
        startfilter=48,
        n_rays=16,
        epochs=400,
        learning_rate=0.0001,
    ):

        self.npz_filename = npz_filename
        self.base_dir = base_dir
        self.downsample_factor = downsample_factor
        self.unet_model_dir = unet_model_dir
        self.star_model_dir = star_model_dir
        self.backbone = backbone
        self.raw_dir = raw_dir
        self.real_mask_dir = real_mask_dir
        self.val_raw_dir = val_raw_dir
        self.val_real_mask_dir = val_real_mask_dir
        self.binary_mask_dir = binary_mask_dir
        self.generate_npz = generate_npz
        self.annisotropy = annisotropy
        self.train_unet = train_unet
        self.train_star = train_star
        self.unet_model_name = unet_model_name
        self.star_model_name = star_model_name
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.depth = depth
        self.n_channel_in = n_channel_in
        self.n_channel_out = n_channel_out
        self.n_rays = n_rays
        self.pattern = pattern
        self.train_loss = train_loss
        self.erosion_iterations = erosion_iterations
        self.kern_size = kern_size
        self.patch_size = patch_size
        self.grid_x = grid_x
        self.grid_y = grid_y
        self.validation_split = validation_split
        self.batch_size = batch_size
        self.use_gpu = use_gpu
        self.search_pattern = "*" + self.pattern
        self.startfilter = startfilter
        self.n_patches_per_image = n_patches_per_image
        self.load_data_sequence = load_data_sequence
        self.val_size = val_size
        self.acceptable_formats = [".tif", ".TIFF", ".TIF", ".png"]
        self.axis_norm = (0, 1, 2)
        self.axes = "ZYXC"
        self.Train()

    class UnetSequencer(Sequence):
        def __init__(
            self,
            filesraw,
            filesmask,
            axis_norm,
            batch_size=1,
            shape=(16, 256, 256),
        ):
            super().__init__()

            self.filesraw = filesraw
            self.filesmask = filesmask
            self.axis_norm = axis_norm
            self.batch_size = batch_size
            self.shape = shape

        def __len__(self):
            return len(self.filesraw) // self.batch_size

        def __getitem__(self, idx):

            batch_x = self.filesraw[
                idx * self.batch_size : (idx + 1) * self.batch_size
            ]
            batch_y = self.filesmask[
                idx * self.batch_size : (idx + 1) * self.batch_size
            ]
            rawlist = []
            masklist = []

            for fname in batch_x:
                raw = read_float(fname)
                if raw.shape == self.shape:
                    raw = normalize(raw, 1, 99.8, axis=self.axis_norm)
                    rawlist.append(raw)
            for fname in batch_y:
                mask = read_int(fname)
                if mask.shape == self.shape:
                    mask = mask > 0
                    mask = binary_fill_holes(mask)
                    masklist.append(mask)

            return np.asarray(rawlist, dtype=np.float32), np.asarray(
                masklist, dtype=np.float32
            )

    class DataSequencer(Sequence):
        def __init__(
            self,
            files,
            axis_norm,
            normalize=True,
            label_me=False,
            binary_me=False,
        ):
            super().__init__()

            self.files = files

            self.axis_norm = axis_norm
            self.label_me = label_me
            self.binary_me = binary_me
            self.normalize = normalize

        def __len__(self):
            return len(self.files)

        def __getitem__(self, i):

            # Read raw images
            if self.normalize is True:
                x = read_float(self.files[i])
                x = normalize(x, 1, 99.8, axis=self.axis_norm)
            if self.label_me is True:
                # Read Label images
                x = read_int(self.files[i])
                if self.binary_me:
                    x = x > 0
                x = x.astype(np.uint16)
            return x

    def Train(self):

        raw_path = os.path.join(self.base_dir, self.raw_dir)
        raw = os.listdir(raw_path)
        if self.load_data_sequence:
            val_raw_path = os.path.join(self.base_dir, self.val_raw_dir)
            val_raw = os.listdir(val_raw_path)
            val_real_mask_path = os.path.join(
                self.base_dir, self.val_real_mask_dir
            )

            Path(val_real_mask_path).mkdir(exist_ok=True)
            val_real_mask = os.listdir(val_real_mask_path)

        mask_path = os.path.join(self.base_dir, self.binary_mask_dir)
        Path(mask_path).mkdir(exist_ok=True)
        mask = os.listdir(mask_path)

        real_mask_path = os.path.join(self.base_dir, self.real_mask_dir)
        Path(real_mask_path).mkdir(exist_ok=True)
        real_mask = os.listdir(real_mask_path)

        print("Instance segmentation masks:", len(real_mask))
        print("Semantic segmentation masks:", len(mask))
        if self.train_star and len(mask) > 0 and len(real_mask) < len(mask):

            print("Making labels")
            mask = sorted(
                glob.glob(
                    self.base_dir + self.binary_mask_dir + "*" + self.pattern
                )
            )

            for fname in mask:
                if any(fname.endswith(f) for f in self.acceptable_formats):
                    image = imread(os.path.join(self.base_dir, fname))

                    Name = os.path.basename(os.path.splitext(fname)[0])
                    if np.max(image) == 1:
                        image = image * 255
                    binary_image = label(image)

                    imwrite(
                        (
                            os.path.join(
                                self.base_dir,
                                self.real_mask_dir + Name + self.pattern,
                            )
                        ),
                        binary_image.astype("uint16"),
                    )

        if len(real_mask) > 0 and len(mask) < len(real_mask):
            print("Generating Binary images")

            real_files_mask = os.listdir(real_mask_path)

            for fname in real_files_mask:
                if any(fname.endswith(f) for f in self.acceptable_formats):
                    image = imread(os.path.join(real_mask_path, fname))
                    if self.erosion_iterations > 0:
                        image = erode_labels(
                            image.astype("uint16"), self.erosion_iterations
                        )
                    Name = os.path.basename(os.path.splitext(fname)[0])

                    binary_image = image > 0

                    imwrite(
                        (
                            os.path.join(
                                self.base_dir,
                                self.binary_mask_dir + Name + self.pattern,
                            )
                        ),
                        binary_image.astype("uint16"),
                    )

        if self.generate_npz:

            raw_data = RawData.from_folder(
                basepath=self.base_dir,
                source_dirs=[self.raw_dir],
                target_dir=self.binary_mask_dir,
                pattern=self.search_pattern,
                axes="ZYX",
            )

            X, Y, XY_axes = create_patches(
                raw_data=raw_data,
                patch_size=self.patch_size,
                n_patches_per_image=self.n_patches_per_image,
                save_file=os.path.join(
                    self.base_dir, self.npz_filename + ".npz"
                ),
            )

        # Training UNET model
        if self.train_unet:
            print("Training UNET model")

            if not self.load_data_sequence:
                load_path = os.path.join(
                    self.base_dir, self.npz_filename + ".npz"
                )
                (X, Y), (X_val, Y_val), axes = load_training_data(
                    load_path,
                    validation_split=self.validation_split,
                    verbose=True,
                )
                self.axes = axes
            else:
                raw_path_list = []
                for fname in raw:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        raw_path_list.append(os.path.join(raw_path, fname))
                val_raw_path_list = []
                for fname in val_raw:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        val_raw_path_list.append(
                            os.path.join(val_raw_path, fname)
                        )

                mask_path_list = []
                for fname in mask:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        mask_path_list.append(os.path.join(mask_path, fname))
                val_real_mask_path_list = []
                for fname in val_real_mask:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        val_real_mask_path_list.append(
                            os.path.join(val_real_mask_path, fname)
                        )

                XY = self.UnetSequencer(
                    raw_path_list,
                    mask_path_list,
                    self.axis_norm,
                    self.batch_size,
                    self.patch_size,
                )

                XY_val = self.UnetSequencer(
                    val_raw_path_list,
                    val_real_mask_path_list,
                    self.axis_norm,
                    self.batch_size,
                    self.patch_size,
                )

            config = Config(
                self.axes,
                self.n_channel_in,
                self.n_channel_out,
                unet_n_depth=self.depth,
                train_epochs=self.epochs,
                train_batch_size=self.batch_size,
                unet_n_first=self.startfilter,
                train_loss=self.train_loss,
                unet_kern_size=self.kern_size,
                train_learning_rate=self.learning_rate,
                train_reduce_lr={"patience": 5, "factor": 0.5},
            )
            print(config)
            vars(config)

            model = CARE(
                config, name=self.unet_model_name, basedir=self.unet_model_dir
            )

            if os.path.exists(
                os.path.join(
                    self.unet_model_dir,
                    os.path.join(self.unet_model_name, "weights_now.h5"),
                )
            ):
                print("Loading checkpoint model")
                model.load_weights(
                    os.path.join(
                        self.unet_model_dir,
                        os.path.join(self.unet_model_name, "weights_now.h5"),
                    )
                )

            if os.path.exists(
                os.path.join(
                    self.unet_model_dir,
                    os.path.join(self.unet_model_name, "weights_last.h5"),
                )
            ):
                print("Loading checkpoint model")
                model.load_weights(
                    os.path.join(
                        self.unet_model_dir,
                        os.path.join(self.unet_model_name, "weights_last.h5"),
                    )
                )

            if os.path.exists(
                os.path.join(
                    self.unet_model_dir,
                    os.path.join(self.unet_model_name, "weights_best.h5"),
                )
            ):
                print("Loading checkpoint model")
                model.load_weights(
                    os.path.join(
                        self.unet_model_dir,
                        os.path.join(self.unet_model_name, "weights_best.h5"),
                    )
                )
            if not self.load_data_sequence:
                history = model.train(
                    X,
                    Y,
                    validation_data=(X_val, Y_val),
                    load_data_sequence=self.load_data_sequence,
                )
            else:
                history = model.train(
                    XY,
                    validation_data=XY_val,
                    load_data_sequence=self.load_data_sequence,
                )

            plt.figure(figsize=(16, 5))
            plot_history(
                history,
                ["loss", "val_loss"],
                ["mse", "val_mse", "mae", "val_mae"],
            )

            print(sorted(list(history.history.keys())))

        if self.train_star:
            print(
                "Training StarDistModel model with", self.backbone, "backbone"
            )

            real_files_mask = os.listdir(real_mask_path)
            if not self.load_data_sequence:
                rng = np.random.RandomState(len(raw) // 2)
                ind = rng.permutation(len(raw))
                self.Y = []
                self.X = []
                for fname in real_files_mask:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        self.Y.append(
                            read_int(os.path.join(real_mask_path, fname))
                        )

                for fname in raw:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        self.X.append(
                            read_float(os.path.join(raw_path, fname))
                        )

                n_val = max(1, int(round(self.validation_split * len(ind))))
                ind_train, ind_val = ind[:-n_val], ind[-n_val:]

                self.X_val, self.Y_val = [self.X[i] for i in ind_val], [
                    self.Y[i] for i in ind_val
                ]
                self.X_trn, self.Y_trn = [self.X[i] for i in ind_train], [
                    self.Y[i] for i in ind_train
                ]

                print("number of images: %3d" % len(self.X))
                print("- training:       %3d" % len(self.X_trn))
                print("- validation:     %3d" % len(self.X_val))
                print(Config3D.__doc__)

                extents = calculate_extents(self.Y_trn)
                self.annisotropy = tuple(np.max(extents) / extents)
                rays = Rays_GoldenSpiral(
                    self.n_rays, anisotropy=self.annisotropy
                )
            if self.load_data_sequence:
                rays = self.n_rays
                self.annisotropy = None
                raw_path_list = []
                for fname in raw:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        raw_path_list.append(os.path.join(raw_path, fname))
                val_raw_path_list = []
                for fname in val_raw:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        val_raw_path_list.append(
                            os.path.join(val_raw_path, fname)
                        )

                self.X_trn = self.DataSequencer(
                    raw_path_list,
                    self.axis_norm,
                    normalize=True,
                    label_me=False,
                )
                real_mask_path_list = []
                for fname in real_mask:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        real_mask_path_list.append(
                            os.path.join(real_mask_path, fname)
                        )
                val_real_mask_path_list = []
                for fname in val_real_mask:
                    if any(fname.endswith(f) for f in self.acceptable_formats):
                        val_real_mask_path_list.append(
                            os.path.join(val_real_mask_path, fname)
                        )

                self.Y_trn = self.DataSequencer(
                    real_mask_path_list,
                    self.axis_norm,
                    normalize=False,
                    label_me=True,
                )

                self.X_val = self.DataSequencer(
                    val_raw_path_list,
                    self.axis_norm,
                    normalize=True,
                    label_me=False,
                )
                self.Y_val = self.DataSequencer(
                    val_real_mask_path_list,
                    self.axis_norm,
                    normalize=False,
                    label_me=True,
                )
                self.train_sample_cache = False

            if self.backbone == "resnet":

                conf = Config3D(
                    rays=rays,
                    anisotropy=self.annisotropy,
                    backbone=self.backbone,
                    train_epochs=self.epochs,
                    train_learning_rate=self.learning_rate,
                    resnet_n_blocks=self.depth,
                    train_checkpoint=self.star_model_dir
                    + self.star_model_name
                    + ".h5",
                    resnet_kernel_size=(
                        self.kern_size,
                        self.kern_size,
                        self.kern_size,
                    ),
                    train_patch_size=self.patch_size,
                    train_batch_size=self.batch_size,
                    resnet_n_filter_base=self.startfilter,
                    train_dist_loss="mse",
                    grid=(1, self.grid_y, self.grid_x),
                    use_gpu=self.use_gpu,
                    n_channel_in=self.n_channel_in,
                )

            if self.backbone == "unet":

                conf = Config3D(
                    rays=rays,
                    anisotropy=self.annisotropy,
                    backbone=self.backbone,
                    train_epochs=self.epochs,
                    train_learning_rate=self.learning_rate,
                    unet_n_depth=self.depth,
                    train_checkpoint=self.star_model_dir
                    + self.star_model_name
                    + ".h5",
                    unet_kernel_size=(
                        self.kern_size,
                        self.kern_size,
                        self.kern_size,
                    ),
                    train_patch_size=self.patch_size,
                    train_batch_size=self.batch_size,
                    unet_n_filter_base=self.startfilter,
                    train_dist_loss="mse",
                    grid=(1, self.grid_y, self.grid_x),
                    use_gpu=self.use_gpu,
                    n_channel_in=self.n_channel_in,
                    train_sample_cache=False,
                )

            print(conf)
            vars(conf)

            Starmodel = StarDist3D(
                conf, name=self.star_model_name, basedir=self.star_model_dir
            )
            print(
                Starmodel._axes_tile_overlap("ZYX"),
                os.path.exists(
                    os.path.join(
                        self.star_model_dir,
                        os.path.join(self.star_model_name, "weights_now.h5"),
                    )
                ),
            )

            if os.path.exists(
                os.path.join(
                    self.star_model_dir,
                    os.path.join(self.star_model_name, "weights_now.h5"),
                )
            ):
                print("Loading checkpoint model")
                Starmodel.load_weights(
                    os.path.join(
                        self.star_model_dir,
                        os.path.join(self.star_model_name, "weights_now.h5"),
                    )
                )

            if os.path.exists(
                os.path.join(
                    self.star_model_dir,
                    os.path.join(self.star_model_name, "weights_last.h5"),
                )
            ):
                print("Loading checkpoint model")
                Starmodel.load_weights(
                    os.path.join(
                        self.star_model_dir,
                        os.path.join(self.star_model_name, "weights_last.h5"),
                    )
                )

            if os.path.exists(
                os.path.join(
                    self.star_model_dir,
                    os.path.join(self.star_model_name, "weights_best.h5"),
                )
            ):
                print("Loading checkpoint model")
                Starmodel.load_weights(
                    os.path.join(
                        self.star_model_dir,
                        os.path.join(self.star_model_name, "weights_best.h5"),
                    )
                )

            historyStar = Starmodel.train(
                self.X_trn,
                self.Y_trn,
                validation_data=(self.X_val, self.Y_val),
                epochs=self.epochs,
            )
            print(sorted(list(historyStar.history.keys())))
            plt.figure(figsize=(16, 5))
            plot_history(
                historyStar,
                ["loss", "val_loss"],
                [
                    "dist_relevant_mae",
                    "val_dist_relevant_mae",
                    "dist_relevant_mse",
                    "val_dist_relevant_mse",
                ],
            )


def read_float(fname):

    return imread(fname).astype("float32")


def read_int(fname):

    return imread(fname).astype("uint16")


def DownsampleData(image, downsample_factor):

    scale_percent = int(100 / downsample_factor)  # percent of original size
    width = int(image.shape[2] * scale_percent / 100)
    height = int(image.shape[1] * scale_percent / 100)
    dim = (width, height)
    smallimage = np.zeros([image.shape[0], height, width])
    for i in range(0, image.shape[0]):
        # resize image
        smallimage[i, :] = zoom(image[i, :].astype("float32"), dim)

    return smallimage



# -*- python -*-
#
#       spatial_image: spatial nd images
#
#       Copyright 2006 INRIA - CIRAD - INRA
#
#       File author(s): Jerome Chopard <jerome.chopard@sophia.inria.fr>
#
#       Distributed under the Cecill-C License.
#       See accompanying file LICENSE.txt or copy at
#           http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html
#
#       OpenAlea WebSite : http://openalea.gforge.inria.fr
#
"""
This module create the main |SpatialImage| object
"""

__license__ = "Cecill-C"
__revision__ = " $Id: $ "

import numpy as np
from scipy import ndimage

# -- deprecation messages --
# import warnings, exceptions
# msg = "SpatialImage.resolution is deprecated, use SpatialImage.voxelsize"
# rezexc = exceptions.PendingDeprecationWarning(msg)


class SpatialImage(np.ndarray):
    """
    Associate meta data to np.ndarray
    """

    def __new__(
        cls, input_array, voxelsize=None, vdim=None, info=None, dtype=None, **kwargs
    ):
        """Instantiate a new |SpatialImage|

        if voxelsize is None, vdim will be used to infer space size and affect
        a voxelsize of 1 in each direction of space

        .. warning :: `resolution` keyword is deprecated. Use `voxelsize` instead.

        :Parameters:
         - `cls` - internal python
         - `input_array` (array) - data to put in the image
         - `voxelsize` (tuple of float) - spatial extension in each direction
                                           of space
         - `vdim` (int) - size of data if vector data are used
         - `info` (dict of str|any) - metainfo
        """
        # if the input_array is 2D we can reshape it to 3D.
        # ~ if input_array.ndim == 2: # Jonathan
        # ~ input_array = input_array.reshape( input_array.shape+(1,) ) # Jonathan

        # initialize datas. For some obscure reason, we want the data
        # to be F-Contiguous in the NUMPY sense. I mean, if this is not
        # respected, we will have problems when communicating with
        # C-Code... yeah, that makes so much sense (fortran-contiguous
        # to be c-readable...).
        dtype = dtype if dtype is not None else input_array.dtype
        if input_array.flags.f_contiguous:
            obj = np.asarray(input_array, dtype=dtype).view(cls)
        else:
            obj = np.asarray(input_array, dtype=dtype, order="F").view(cls)

        voxelsize = kwargs.get("resolution", voxelsize)  # to manage transition

        if voxelsize is None:
            # ~ voxelsize = (1.,) * 3
            voxelsize = (1.0,) * input_array.ndim  # Jonathan
        else:
            # ~ if len(voxelsize) != 3 :
            if (input_array.ndim != 4) and (
                len(voxelsize) != input_array.ndim
            ):  # Jonathan _ Compatibility with "champs_*.inr.gz" generated by Baloo & SuperBaloo
                raise ValueError("data dimension and voxelsize mismatch")

        obj.voxelsize = tuple(voxelsize)
        obj.vdim = vdim if vdim else 1

        # set metadata
        if info is None:
            obj.info = {}
        else:
            obj.info = dict(info)

        # return
        return obj

    def _get_resolution(self):
        # warnings.warn(rezexc)
        return self.voxelsize

    def _set_resolution(self, val):
        # warnings.warn(rezexc)
        self.voxelsize = val

    resolution = property(_get_resolution, _set_resolution)

    @property
    def real_shape(self):
        # ~ return np.multiply(self.shape[:3], self.voxelsize)
        return np.multiply(self.shape, self.voxelsize)  # Jonathan

    def invert_z_axis(self):
        """
        invert allong 'Z' axis
        """
        self = self[:, :, ::-1]

    def __array_finalize__(self, obj):
        if obj is None:
            return

        # assert resolution
        res = getattr(obj, "voxelsize", None)
        if res is None:  # assert vdim == 1
            res = (1.0,) * len(obj.shape)

        self.voxelsize = tuple(res)

        # metadata
        self.info = dict(getattr(obj, "info", {}))

    def clone(self, data):
        """Clone the current image metadata
        on the given data.
        .. warning:: vdim is defined according to self.voxelsize and data.shape
        :Parameters:
         - `data` - (array)

        :Returns Type: |SpatialImage|
        """
        if len(data.shape) == len(self.voxelsize):
            vdim = 1
        elif len(data.shape) - len(self.voxelsize) == 1:
            vdim = data.shape[-1]
        else:
            raise UserWarning("unable to handle such data dimension")

        return SpatialImage(data, self.voxelsize, vdim, self.info)

    @classmethod
    def valid_array(cls, array_like):
        return (
            isinstance(array_like, (np.ndarray, cls)) and array_like.flags.f_contiguous
        )


def empty_image_like(spatial_image):
    array = np.zeros(spatial_image.shape, dtype=spatial_image.dtype)
    return SpatialImage(array, spatial_image.voxelsize, vdim=1)


def null_vector_field_like(spatial_image):
    array = np.zeros(list(spatial_image.shape) + [3], dtype=np.float32)
    return SpatialImage(array, spatial_image.voxelsize, vdim=3)


def random_vector_field_like(spatial_image, smooth=0, max_=1):
    # ~ if spatial_image.vdim == 1:
    # ~ shape = spatial_image.shape+(3,)
    # ~ else:
    # ~ shape = spatial_image.shape
    shape = spatial_image.shape  # Jonathan
    array = np.random.uniform(-max_, max_, shape)
    if smooth:
        array = ndimage.gaussian_filter(array, smooth)
    return SpatialImage(array, spatial_image.voxelsize, dtype=np.float32)


def checkerboard(nx=9, ny=8, nz=5, size=10, vs=(1.0, 1.0, 1.0), dtype=np.uint8):
    """Creates a 3D checkerboard image with `nx` squares in width,
    `ny` squares in height and `nz` squares in depth. The length of the edge in real units
    of each square is `size`."""

    sxv, syv, szv = np.array([size] * 3) / np.array(vs)
    array = np.zeros((sxv * nx, syv * ny, szv * nz), dtype=dtype, order="F")
    typeinfo = np.iinfo(dtype)

    # -- wooo surely not the most beautiful implementation out here --
    for k in range(nz):
        kval = typeinfo.max if (k % 2 == 0) else typeinfo.min
        jval = kval
        for j in range(ny):
            ival = jval
            for i in range(nx):
                array[
                    i * sxv : i * sxv + sxv,
                    j * syv : j * syv + syv,
                    k * szv : k * szv + szv,
                ] = ival
                ival = typeinfo.max if (ival == typeinfo.min) else typeinfo.min
            jval = typeinfo.max if (jval == typeinfo.min) else typeinfo.min
        kval = typeinfo.max if (kval == typeinfo.min) else typeinfo.min

    return SpatialImage(array, vs, dtype=dtype)


def is2D(image):
    """
    Test if the `image` (array) is in 2D or 3D.
    Return True if 2D, False if not.
    """
    if len(image.shape) == 2 or image.shape[2] == 1:
        return True
    else:
        return False



class UnetStarMask:
    def __init__(self, boxA, cordB):

        self.boxA = boxA
        self.cordB = cordB

    def semi_masking(self):

        self.masksemiD()
        return self.semiinclude

    def masking(self):

        self.masknD()

        return self.include

    def masknD(self):

        self.ndim = len(self.cordB)

        self.include = False
        includelist = [self.Conditioncheck(p) for p in range(0, self.ndim)]
        if True in includelist:
            self.include = True

    def Conditioncheck(self, p):

        include = True

        if self.cordB[p] >= self.boxA[p] and self.cordB[p] <= self.boxA[p + self.ndim]:

            include = False

        return include

    def masksemiD(self):

        self.ndim = len(self.cordB)

        self.semiinclude = False
        includelist = [self.Conditioncheck(p) for p in range(1, self.ndim)]
        if True in includelist:
            self.semiinclude = True



from skimage import measure
import numpy as np
from skimage.util import map_array
import vollseg.utils


class NMSLabel:
    def __init__(self, image, nms_thresh, z_thresh=1):
        self.image = image
        self.nms_thresh = nms_thresh
        self.z_thresh = z_thresh

    def supresslabels(self):

        print("Supressing spurious labels, this can take some time")
        properties = measure.regionprops(self.image)
        Bbox = [prop.bbox for prop in properties]
        Labels = [prop.label for prop in properties]
        self.supresslabel = {}
        while len(Labels) > 0:
            last = len(Labels) - 1
            i = Labels[last]
            suppress = [last]
            for pos in range(0, last):
                # grab the current index
                j = Labels[pos]
                self.iou(Bbox[last], Bbox[pos], i, j)

            Labels = np.delete(Labels, suppress)

        for (k, v) in self.supresslabel.items():
            pixel_condition = self.image == k
            pixel_replace_condition = v
            self.image = vollseg.utils.image_conditionals(
                self.image, pixel_condition, pixel_replace_condition
            )

        return self.image

    def supressregions(self):

        print("Supressing spurious regions, this can take some time")
        properties = measure.regionprops(self.image)
        Bbox = [prop.bbox for prop in properties]
        Labels = [prop.label for prop in properties]

        self.originallabels = []
        self.newlabels = []
        for pos in range(len(Labels)):
            current_label = Labels[pos]
            self.smallz(Bbox[pos], current_label)

        if len(self.originallabels) > 0:
            relabeled = map_array(
                self.image, np.asarray(self.originallabels), np.asarray(self.newlabels)
            )
        else:
            relabeled = self.image
        return relabeled

    def smallz(self, box, label):

        ndim = len(self.image.shape)
        if ndim == 3:
            z = abs(box[0] - box[3])
            if z < self.z_thresh:
                self.originallabels.append(label)
                self.newlabels.append(0)
            else:
                self.originallabels.append(label)
                self.newlabels.append(label)

    def iou(self, boxA, boxB, labelA, labelB):

        ndim = len(self.image.shape)

        if ndim == 2:
            xA = max(boxA[0], boxB[0])
            yA = max(boxA[1], boxB[1])
            xB = min(boxA[2], boxB[2])
            yB = min(boxA[3], boxB[3])

            # BoxA contains BoxB
            if (
                boxA[0] <= boxB[0]
                and boxA[2] >= boxB[2]
                and boxA[1] <= boxB[1]
                and boxA[3] >= boxB[3]
            ):
                self.supresslabel[labelB] = labelA
            # BoxB contains BoxA
            if (
                boxB[0] <= boxA[0]
                and boxB[2] >= boxA[2]
                and boxB[1] <= boxA[1]
                and boxB[3] >= boxA[3]
            ):
                self.supresslabel[labelA] = labelB
            else:
                # compute the area of intersection rectangle
                interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)
                # compute the area of both the prediction and ground-truth
                # rectangles
                boxAArea = (boxA[2] - boxA[0] + 1) * (boxA[3] - boxA[1] + 1)
                boxBArea = (boxB[2] - boxB[0] + 1) * (boxB[3] - boxB[1] + 1)
                # compute the intersection over union by taking the intersection
                # area and dividing it by the sum of prediction + ground-truth
                # areas - the interesection area
                iou = interArea / float(boxAArea + boxBArea - interArea)
                if iou >= self.nms_thresh:
                    self.supresslabel[labelA] = labelB

        if ndim == 3:

            zA = max(boxA[0], boxB[0])
            yA = max(boxA[1], boxB[1])
            xA = max(boxA[2], boxB[2])
            zB = min(boxA[3], boxB[3])
            yB = min(boxA[4], boxB[4])
            xB = min(boxA[5], boxB[5])
            if (
                boxA[1] <= boxB[1]
                and boxA[4] >= boxB[4]
                and boxA[2] <= boxB[2]
                and boxA[5] >= boxB[5]
            ):
                self.supresslabel[labelB] = labelA
            elif (
                boxB[1] <= boxA[1]
                and boxB[4] >= boxA[4]
                and boxB[2] <= boxA[2]
                and boxB[5] >= boxA[5]
            ):
                self.supresslabel[labelA] = labelB

            else:
                # compute the area of intersection rectangle
                interArea = (
                    max(0, xB - xA + 1) * max(0, yB - yA + 1) * max(0, zB - zA + 1)
                )
                # compute the area of both the prediction and ground-truth
                # rectangles
                boxAArea = (
                    (boxA[3] - boxA[0] + 1)
                    * (boxA[4] - boxA[1] + 1)
                    * (boxA[5] - boxA[2] + 1)
                )
                boxBArea = (
                    (boxB[3] - boxB[0] + 1)
                    * (boxB[4] - boxB[1] + 1)
                    * (boxB[5] - boxB[2] + 1)
                )
                # compute the intersection over union by taking the intersection
                # area and dividing it by the sum of prediction + ground-truth
                # areas - the interesection area
                iou = interArea / float(boxAArea + boxBArea - interArea)

                if iou >= self.nms_thresh:
                    self.supresslabel[labelA] = labelB



"""
# 3D Cellpose Extension.
# Copyright (C) 2021 D. Eschweiler, J. Stegmaier
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the Liceense at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Please refer to the documentation for more information about the software
# as well as for installation instructions.
"""
# Modified by Varun Kapoor

import concurrent
import glob
import os
from csbdeep.utils import normalize
import h5py
import numpy as np
from scipy.ndimage import distance_transform_edt, generic_filter, zoom
from scipy.spatial import ConvexHull, Delaunay
from skimage import filters, io, measure, morphology
from tqdm import tqdm
from pathlib import Path
import csv
import json


def load_json(fpath):
    with open(fpath) as f:
        return json.load(f)


def save_json(data, fpath, **kwargs):
    with open(fpath, "w") as f:
        f.write(json.dumps(data, **kwargs))


def create_csv(
    data_list,
    save_path,
    save_train="_train.csv",
    save_test="_test.csv",
    save_val="_val.csv",
    test_split=0.1,
    val_split=0.1,
    shuffle=True,
):

    if shuffle:
        np.random.shuffle(data_list)

    # Get number of files for each split
    num_files = len(data_list)
    num_test_files = int(test_split * num_files)
    num_val_files = int((num_files - num_test_files) * val_split)
    num_train_files = num_files - num_test_files - num_val_files

    # Get file indices
    file_idx = np.arange(num_files)

    # Save csv files
    if num_test_files > 0:
        test_idx = sorted(
            np.random.choice(file_idx, size=num_test_files, replace=False)
        )
        with open(save_path + save_test, "w") as fh:
            writer = csv.writer(fh, delimiter=";")
            for idx in test_idx:
                writer.writerow(data_list[idx])
    else:
        test_idx = []

    if num_val_files > 0:
        val_idx = sorted(
            np.random.choice(
                list(set(file_idx) - set(test_idx)), size=num_val_files, replace=False
            )
        )
        with open(save_path + save_val, "w") as fh:
            writer = csv.writer(fh, delimiter=";")
            for idx in val_idx:
                writer.writerow(data_list[idx])
    else:
        val_idx = []

    if num_train_files > 0:
        train_idx = sorted(list(set(file_idx) - set(test_idx) - set(val_idx)))
        with open(save_path + save_train, "w") as fh:
            writer = csv.writer(fh, delimiter=";")
            for idx in train_idx:
                writer.writerow(data_list[idx])


def _h5_writer(data_list, save_path, group_root="data", group_names=["image"]):
    save_path = os.path.abspath(save_path)

    assert len(data_list) == len(group_names), "Each data matrix needs a group name"

    with h5py.File(save_path, "w") as f_handle:
        grp = f_handle.create_group(group_root)
        for data, group_name in zip(data_list, group_names):
            grp.create_dataset(group_name, data=data, chunks=True, compression="gzip")


def _flood_fill_hull(image):
    points = np.transpose(np.where(image))
    hull = ConvexHull(points)
    deln = Delaunay(points[hull.vertices])
    idx = np.stack(np.indices(image.shape), axis=-1)
    out_idx = np.nonzero(deln.find_simplex(idx) + 1)
    out_img = np.zeros(image.shape)
    out_img[out_idx] = 1

    return out_img


def _calculate_flows(instance_mask, bg_label=0):
    flow_x = np.zeros(instance_mask.shape, dtype=np.float32)
    flow_y = np.zeros(instance_mask.shape, dtype=np.float32)
    flow_z = np.zeros(instance_mask.shape, dtype=np.float32)
    regions = measure.regionprops(instance_mask)

    futures = []

    with concurrent.futures.ThreadPoolExecutor(max_workers=os.cpu_count()) as executor:
        for props in regions:
            if props.label != bg_label:
                futures.append(executor.submit(_compute_flow, props, instance_mask))

        for r in concurrent.futures.as_completed(futures):
            coords, c = r.result()
            flow_x[coords] = np.tanh((coords[0] - c[0]) / 5)
            flow_y[coords] = np.tanh((coords[1] - c[1]) / 5)
            flow_z[coords] = np.tanh((coords[2] - c[2]) / 5)

    return flow_x, flow_y, flow_z


def _compute_flow(props, instance_mask):
    c = props.centroid
    coords = np.where(instance_mask == props.label)
    return coords, c


def rescale_data(data, zoom_factor, order=0):
    if any([zf != 1 for zf in zoom_factor]):
        data_shape = data.shape
        data = zoom(data, zoom_factor, order=3)
        print(f"Rescaled image from size {data_shape} to {data.shape}")

    return data


def foreground_from_mip(img):
    mip = np.ones_like(img)

    for ndim in range(img.ndim):
        mip_tmp = np.max(img, axis=ndim)
        mip_tmp = mip_tmp > 0.05
        mip_tmp = np.expand_dims(mip_tmp, ndim)
        mip_tmp = np.repeat(mip_tmp, img.shape[ndim], ndim)

        mip *= mip_tmp

    return mip


def prepare_images(
    data_path="",
    save_path="",
    folders=[""],
    identifier="*.tif",
    axis_norm=(0, 1, 2),
    get_distance=False,
    get_illumination=False,
    get_variance=False,
    variance_size=(5, 5, 5),
    fg_footprint_size=5,
    channel=0,
):
    data_path = os.path.abspath(data_path)
    save_path = os.path.abspath(save_path)
    Path(save_path).mkdir(exist_ok=True)
    for image_folder in folders:
        image_list = glob.glob(os.path.join(data_path, image_folder, identifier))
        for file in tqdm(image_list):

            # load the image

            processed_img = io.imread(file)
            processed_img = processed_img.astype(np.float32)

            # get the desired channel, if the image is a multichannel image
            if processed_img.ndim == 4:
                processed_img = processed_img[..., channel]

            # normalize the image

            processed_img = normalize(processed_img, 1, 99.8, axis=axis_norm)
            processed_img = processed_img.astype(np.float32)

            save_imgs = [
                processed_img,
            ]
            save_groups = [
                "image",
            ]

            if get_illumination:
                print("Extracting illumination image...")

                # create downscales image for computantially intensive processing
                small_img = processed_img

                # create an illuminance image (downscale for faster processing)
                illu_img = morphology.closing(small_img, footprint=morphology.ball(7))
                illu_img = filters.gaussian(illu_img, 2).astype(np.float32)

                # rescale illuminance image
                illu_img = np.repeat(illu_img, 2, axis=0)
                illu_img = np.repeat(illu_img, 2, axis=1)
                illu_img = np.repeat(illu_img, 2, axis=2)
                dim_missmatch = np.array(processed_img.shape) - np.array(illu_img.shape)
                if dim_missmatch[0] < 0:
                    illu_img = illu_img[: dim_missmatch[0], ...]
                if dim_missmatch[1] < 0:
                    illu_img = illu_img[:, : dim_missmatch[1], :]
                if dim_missmatch[2] < 0:
                    illu_img = illu_img[..., : dim_missmatch[2]]

                save_imgs.append(illu_img.astype(np.float32))
                save_groups.append("illumination")

            if get_distance:
                print("Extracting distance image...")

                # create downscales image for computantially intensive processing
                small_img = processed_img

                # find suitable threshold
                thresh = filters.threshold_otsu(small_img)
                fg_img = small_img > thresh

                # remove noise and fill holes
                fg_img = morphology.binary_closing(
                    fg_img, footprint=morphology.ball(fg_footprint_size)
                )
                fg_img = morphology.binary_opening(
                    fg_img, footprint=morphology.ball(fg_footprint_size)
                )
                fg_img = _flood_fill_hull(fg_img)
                fg_img = fg_img.astype(np.bool)

                # create distance transform
                fg_img = distance_transform_edt(fg_img) - distance_transform_edt(
                    ~fg_img
                )

                # rescale distance image
                fg_img = np.repeat(fg_img, 4, axis=0)
                fg_img = np.repeat(fg_img, 4, axis=1)
                fg_img = np.repeat(fg_img, 4, axis=2)
                dim_missmatch = np.array(processed_img.shape) - np.array(fg_img.shape)
                if dim_missmatch[0] < 0:
                    fg_img = fg_img[: dim_missmatch[0], ...]
                if dim_missmatch[1] < 0:
                    fg_img = fg_img[:, : dim_missmatch[1], :]
                if dim_missmatch[2] < 0:
                    fg_img = fg_img[..., : dim_missmatch[2]]

                save_imgs.append(fg_img.astype(np.float32))
                save_groups.append("distance")

            if get_variance:
                print("Extracting variance image...")

                # create downscales image for computantially intensive processing
                small_img = processed_img

                # create variance image
                std_img = generic_filter(small_img, np.std, size=variance_size)

                # rescale variance image
                std_img = np.repeat(std_img, 4, axis=0)
                std_img = np.repeat(std_img, 4, axis=1)
                std_img = np.repeat(std_img, 4, axis=2)
                dim_missmatch = np.array(processed_img.shape) - np.array(std_img.shape)
                if dim_missmatch[0] < 0:
                    std_img = std_img[: dim_missmatch[0], ...]
                if dim_missmatch[1] < 0:
                    std_img = std_img[:, : dim_missmatch[1], :]
                if dim_missmatch[2] < 0:
                    std_img = std_img[..., : dim_missmatch[2]]

                save_imgs.append(std_img.astype(np.float32))
                save_groups.append("variance")

            # save the data
            save_name = os.path.basename(os.path.splitext(file)[0])
            save_name = os.path.join(save_path, save_name + ".h5")
            _h5_writer(
                save_imgs,
                save_name,
                group_root="data",
                group_names=save_groups,
            )


def prepare_masks(
    data_path="",
    save_path="",
    folders=[""],
    identifier="*.tif",
    bg_label=0,
    get_flows=True,
    get_boundary=False,
    get_seeds=False,
    get_distance=True,
    corrupt_prob=0.0,
    zoom_factor=(1, 1, 1),
):
    data_path = os.path.abspath(data_path)
    save_path = os.path.abspath(save_path)
    Path(save_path).mkdir(exist_ok=True)
    for mask_folder in folders:
        mask_list = glob.glob(os.path.join(data_path, mask_folder, identifier))
        experiment_identifier = (
            "corrupt" + str(corrupt_prob).replace(".", "") if corrupt_prob > 0 else ""
        )
        os.makedirs(
            os.path.join(data_path, mask_folder, experiment_identifier),
            exist_ok=True,
        )
        for file in tqdm(mask_list):

            # load the mask
            instance_mask = io.imread(file)
            instance_mask = instance_mask.astype(np.uint16)
            instance_mask[instance_mask == bg_label] = 0

            # rescale the mask
            instance_mask = rescale_data(instance_mask, zoom_factor, order=0)

            if corrupt_prob > 0:
                # Randomly merge neighbouring instances
                labels = list(set(np.unique(instance_mask)) - {bg_label})
                instance_mask_eroded = morphology.erosion(
                    instance_mask, footprint=morphology.ball(3)
                )
                instance_mask_dilated = morphology.dilation(
                    instance_mask, footprint=morphology.ball(3)
                )
                for label in labels:
                    if np.random.rand() < corrupt_prob:
                        neighbour_labels = list(
                            instance_mask_eroded[instance_mask == label]
                        ) + list(instance_mask_dilated[instance_mask == label])
                        neighbour_labels = list(set(neighbour_labels) - {label})
                        if len(neighbour_labels) > 0:
                            replace_label = np.random.choice(neighbour_labels)
                            instance_mask[instance_mask == label] = replace_label

            save_groups = [
                "instance",
            ]
            save_masks = [
                instance_mask,
            ]

            # get the boundary mask
            if get_boundary:
                membrane_mask = (
                    morphology.dilation(instance_mask, footprint=morphology.ball(2))
                    - instance_mask
                )
                membrane_mask = membrane_mask != 0
                membrane_mask = membrane_mask.astype(np.float32)
                save_groups.append("boundary")
                save_masks.append(membrane_mask)

            # get the distance mask
            if get_distance:
                fg_img = instance_mask

                fg_img = _flood_fill_hull(fg_img > 0)
                fg_img = fg_img.astype(np.bool)
                distance_mask = distance_transform_edt(fg_img) - distance_transform_edt(
                    ~fg_img
                )
                distance_mask = distance_mask.astype(np.float32)
                distance_mask = np.repeat(distance_mask, 4, axis=0)
                distance_mask = np.repeat(distance_mask, 4, axis=1)
                distance_mask = np.repeat(distance_mask, 4, axis=2)
                dim_missmatch = np.array(instance_mask.shape) - np.array(
                    distance_mask.shape
                )
                if dim_missmatch[0] < 0:
                    distance_mask = distance_mask[: dim_missmatch[0], ...]
                if dim_missmatch[1] < 0:
                    distance_mask = distance_mask[:, : dim_missmatch[1], :]
                if dim_missmatch[2] < 0:
                    distance_mask = distance_mask[..., : dim_missmatch[2]]
                save_groups.append("distance")
                save_masks.append(distance_mask)

            # get the centroid mask
            if get_seeds:
                centroid_mask = np.zeros(instance_mask.shape, dtype=np.float32)
                regions = measure.regionprops(instance_mask)

                for props in regions:
                    if props.label == bg_label:
                        continue

                    c = props.centroid
                    centroid_mask[np.int(c[0]), np.int(c[1]), np.int(c[2])] = 1

                save_groups.append("seeds")
                save_masks.append(centroid_mask)

            # calculate the flow field
            if get_flows:
                flow_x, flow_y, flow_z = _calculate_flows(
                    instance_mask, bg_label=bg_label
                )

                save_groups.extend(["flow_x", "flow_y", "flow_z"])
                save_masks.extend([flow_x, flow_y, flow_z])

            # save the data
            save_name = os.path.basename(os.path.splitext(file)[0])
            save_name = os.path.join(save_path, save_name + ".h5")
            _h5_writer(
                save_masks,
                save_name,
                group_root="data",
                group_names=save_groups,
            )



import torch.nn as nn
import importlib
import logging
import shutil
import sys
import h5py
from torch import optim
from .buildingblocks import (
    DoubleConv,
    ResNetBlock,
    ResNetBlockSE,
    create_decoders,
    create_encoders,
)
import os
import torch
import numpy as np


class AbstractUNet(nn.Module):
    """
    Base class for standard and residual UNet.
    Args:
        in_channels (int): number of input channels
        out_channels (int): number of output segmentation masks;
            Note that of out_channels might correspond to either
            different semantic classes or to different binary segmentation mask.
            It's up to the user of the class to interpret the out_channels and
            use the proper loss criterion during training (i.e. CrossEntropyLoss (multi-class)
            or BCEWithLogitsLoss (two-class) respectively)
        f_maps (int, tuple): number of feature maps at each level of the encoder; if it's an integer the number
            of feature maps is given by the geometric progression: f_maps ^ k, k=1,2,3,4
        final_sigmoid (bool): if True apply element-wise nn.Sigmoid after the final 1x1 convolution,
            otherwise apply nn.Softmax. In effect only if `self.training == False`, i.e. during validation/testing
        basic_module: basic model for the encoder/decoder (DoubleConv, ResNetBlock, ....)
        layer_order (string): determines the order of layers in `SingleConv` module.
            E.g. 'crg' stands for GroupNorm3d+Conv3d+ReLU. See `SingleConv` for more info
        num_groups (int): number of groups for the GroupNorm
        num_levels (int): number of levels in the encoder/decoder path (applied only if f_maps is an int)
            default: 4
        is_segmentation (bool): if True and the model is in eval mode, Sigmoid/Softmax normalization is applied
            after the final convolution; if False (regression problem) the normalization layer is skipped
        conv_kernel_size (int or tuple): size of the convolving kernel in the basic_module
        pool_kernel_size (int or tuple): the size of the window
        conv_padding (int or tuple): add zero-padding added to all three sides of the input
        is3d (bool): if True the model is 3D, otherwise 2D, default: True
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid,
        basic_module,
        f_maps=[32, 64, 128, 256],
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=True,
        conv_kernel_size=3,
        pool_kernel_size=2,
        conv_padding=1,
        is3d=True,
    ):
        super().__init__()
        if isinstance(f_maps, int):
            f_maps = number_of_features_per_level(
                f_maps, num_levels=num_levels
            )

        assert isinstance(f_maps, list) or isinstance(f_maps, tuple)
        assert len(f_maps) > 1, "Required at least 2 levels in the U-Net"
        if "g" in layer_order:
            assert (
                num_groups is not None
            ), "num_groups must be specified if GroupNorm is used"

        # create encoder path
        self.encoders = create_encoders(
            in_channels,
            f_maps,
            basic_module,
            conv_kernel_size,
            conv_padding,
            layer_order,
            num_groups,
            pool_kernel_size,
            is3d,
        )

        # create decoder path
        self.decoders = create_decoders(
            f_maps,
            basic_module,
            conv_kernel_size,
            conv_padding,
            layer_order,
            num_groups,
            is3d,
        )

        # in the last layer a 1×1 convolution reduces the number of output channels to the number of labels
        if is3d:
            self.final_conv = nn.Conv3d(f_maps[0], out_channels, 1)
        else:
            self.final_conv = nn.Conv2d(f_maps[0], out_channels, 1)

        if is_segmentation:
            # semantic segmentation problem
            if final_sigmoid:
                self.final_activation = nn.Sigmoid()
            else:
                self.final_activation = nn.Softmax(dim=1)
        else:
            # regression problem
            self.final_activation = None

    def forward(self, x):
        # encoder part
        encoders_features = []
        for encoder in self.encoders:
            x = encoder(x)
            # reverse the encoder outputs to be aligned with the decoder
            encoders_features.insert(0, x)

        # remove the last encoder's output from the list
        # !!remember: it's the 1st in the list
        encoders_features = encoders_features[1:]

        # decoder part
        for decoder, encoder_features in zip(self.decoders, encoders_features):
            # pass the output from the corresponding encoder and the output
            # of the previous decoder
            x = decoder(encoder_features, x)

        x = self.final_conv(x)

        # apply final_activation (i.e. Sigmoid or Softmax) only during prediction.
        # During training the network outputs logits
        if not self.training and self.final_activation is not None:
            x = self.final_activation(x)

        return x


class UNet3D(AbstractUNet):
    """
    3DUnet model from
    `"3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation"
        <https://arxiv.org/pdf/1606.06650.pdf>`.
    Uses `DoubleConv` as a basic_module and nearest neighbor upsampling in the decoder
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=[32, 64, 128, 256],
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=True,
        conv_padding=1,
        **kwargs,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=DoubleConv,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            conv_padding=conv_padding,
            is3d=True,
        )


class ResidualUNet3D(AbstractUNet):
    """
    Residual 3DUnet model implementation based on https://arxiv.org/pdf/1706.00120.pdf.
    Uses ResNetBlock as a basic building block, summation joining instead
    of concatenation joining and transposed convolutions for upsampling (watch out for block artifacts).
    Since the model effectively becomes a residual net, in theory it allows for deeper UNet.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=[32, 64, 128, 256],
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=True,
        conv_padding=1,
        **kwargs,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=ResNetBlock,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            conv_padding=conv_padding,
            is3d=True,
        )


class ResidualUNetSE3D(AbstractUNet):
    """_summary_
    Residual 3DUnet model implementation with squeeze and excitation based on
    https://arxiv.org/pdf/1706.00120.pdf.
    Uses ResNetBlockSE as a basic building block, summation joining instead
    of concatenation joining and transposed convolutions for upsampling (watch
    out for block artifacts). Since the model effectively becomes a residual
    net, in theory it allows for deeper UNet.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=[32, 64, 128, 256],
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=True,
        conv_padding=1,
        **kwargs,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=ResNetBlockSE,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            conv_padding=conv_padding,
            is3d=True,
        )


class UNet2D(AbstractUNet):
    """
    2DUnet model from
    `"U-Net: Convolutional Networks for Biomedical Image Segmentation" <https://arxiv.org/abs/1505.04597>`
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=[32, 64, 128, 256],
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=True,
        conv_padding=1,
        **kwargs,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=DoubleConv,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            conv_padding=conv_padding,
            is3d=False,
        )


def get_model(model_config):
    model_class = get_class(
        model_config["name"], modules=["pytorch3dunet.unet3d.model"]
    )
    return model_class(**model_config)


def save_checkpoint(state, is_best, checkpoint_dir):
    """Saves model and training parameters at '{checkpoint_dir}/last_checkpoint.pytorch'.
    If is_best==True saves '{checkpoint_dir}/best_checkpoint.pytorch' as well.
    Args:
        state (dict): contains model's state_dict, optimizer's state_dict, epoch
            and best evaluation metric value so far
        is_best (bool): if True state contains the best model seen so far
        checkpoint_dir (string): directory where the checkpoint are to be saved
    """

    if not os.path.exists(checkpoint_dir):
        os.mkdir(checkpoint_dir)

    last_file_path = os.path.join(checkpoint_dir, "last_checkpoint.pytorch")
    torch.save(state, last_file_path)
    if is_best:
        best_file_path = os.path.join(
            checkpoint_dir, "best_checkpoint.pytorch"
        )
        shutil.copyfile(last_file_path, best_file_path)


def load_checkpoint(
    checkpoint_path,
    model,
    optimizer=None,
    model_key="model_state_dict",
    optimizer_key="optimizer_state_dict",
):
    """Loads model and training parameters from a given checkpoint_path
    If optimizer is provided, loads optimizer's state_dict of as well.
    Args:
        checkpoint_path (string): path to the checkpoint to be loaded
        model (torch.nn.Module): model into which the parameters are to be copied
        optimizer (torch.optim.Optimizer) optional: optimizer instance into
            which the parameters are to be copied
    Returns:
        state
    """
    if not os.path.exists(checkpoint_path):
        raise OSError(f"Checkpoint '{checkpoint_path}' does not exist")

    state = torch.load(checkpoint_path, map_location="cpu")
    model.load_state_dict(state[model_key])

    if optimizer is not None:
        optimizer.load_state_dict(state[optimizer_key])

    return state


def save_network_output(output_path, output, logger=None):
    if logger is not None:
        logger.info(f"Saving network output to: {output_path}...")
    output = output.detach().cpu()[0]
    with h5py.File(output_path, "w") as f:
        f.create_dataset("predictions", data=output, compression="gzip")


loggers = {}


def get_logger(name, level=logging.INFO):
    global loggers
    if loggers.get(name) is not None:
        return loggers[name]
    else:
        logger = logging.getLogger(name)
        logger.setLevel(level)
        # Logging to console
        stream_handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            "%(asctime)s [%(threadName)s] %(levelname)s %(name)s - %(message)s"
        )
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

        loggers[name] = logger

        return logger


def get_number_of_learnable_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


class RunningAverage:
    """Computes and stores the average"""

    def __init__(self):
        self.count = 0
        self.sum = 0
        self.avg = 0

    def update(self, value, n=1):
        self.count += n
        self.sum += value * n
        self.avg = self.sum / self.count


def number_of_features_per_level(init_channel_number, num_levels):
    return [init_channel_number * 2**k for k in range(num_levels)]


class _TensorboardFormatter:
    """
    Tensorboard formatters converts a given batch of images (be it input/output to the network or the target segmentation
    image) to a series of images that can be displayed in tensorboard. This is the parent class for all tensorboard
    formatters which ensures that returned images are in the 'CHW' format.
    """

    def __init__(self, **kwargs):
        pass

    def __call__(self, name, batch):
        """
        Transform a batch to a series of tuples of the form (tag, img), where `tag` corresponds to the image tag
        and `img` is the image itself.
        Args:
             name (str): one of 'inputs'/'targets'/'predictions'
             batch (torch.tensor): 4D or 5D torch tensor
        """

        def _check_img(tag_img):
            tag, img = tag_img

            assert (
                img.ndim == 2 or img.ndim == 3
            ), "Only 2D (HW) and 3D (CHW) images are accepted for display"

            if img.ndim == 2:
                img = np.expand_dims(img, axis=0)
            else:
                C = img.shape[0]
                assert (
                    C == 1 or C == 3
                ), "Only (1, H, W) or (3, H, W) images are supported"

            return tag, img

        tagged_images = self.process_batch(name, batch)

        return list(map(_check_img, tagged_images))

    def process_batch(self, name, batch):
        raise NotImplementedError


class DefaultTensorboardFormatter(_TensorboardFormatter):
    def __init__(self, skip_last_target=False, **kwargs):
        super().__init__(**kwargs)
        self.skip_last_target = skip_last_target

    def process_batch(self, name, batch):
        if name == "targets" and self.skip_last_target:
            batch = batch[:, :-1, ...]

        tag_template = "{}/batch_{}/channel_{}/slice_{}"

        tagged_images = []

        if batch.ndim == 5:
            # NCDHW
            slice_idx = batch.shape[2] // 2  # get the middle slice
            for batch_idx in range(batch.shape[0]):
                for channel_idx in range(batch.shape[1]):
                    tag = tag_template.format(
                        name, batch_idx, channel_idx, slice_idx
                    )
                    img = batch[batch_idx, channel_idx, slice_idx, ...]
                    tagged_images.append((tag, self._normalize_img(img)))
        else:
            # batch has no channel dim: NDHW
            slice_idx = batch.shape[1] // 2  # get the middle slice
            for batch_idx in range(batch.shape[0]):
                tag = tag_template.format(name, batch_idx, 0, slice_idx)
                img = batch[batch_idx, slice_idx, ...]
                tagged_images.append((tag, self._normalize_img(img)))

        return tagged_images

    @staticmethod
    def _normalize_img(img):
        return np.nan_to_num((img - np.min(img)) / np.ptp(img))


def _find_masks(batch, min_size=10):
    """Center the z-slice in the 'middle' of a given instance, given a batch of instances
    Args:
        batch (ndarray): 5d numpy tensor (NCDHW)
    """
    result = []
    for b in batch:
        assert b.shape[0] == 1
        patch = b[0]
        z_sum = patch.sum(axis=(1, 2))
        coords = np.where(z_sum > min_size)[0]
        if len(coords) > 0:
            ind = coords[len(coords) // 2]
            result.append(b[:, ind : ind + 1, ...])
        else:
            ind = b.shape[1] // 2
            result.append(b[:, ind : ind + 1, ...])

    return np.stack(result, axis=0)


def get_tensorboard_formatter(formatter_config):
    if formatter_config is None:
        return DefaultTensorboardFormatter()

    class_name = formatter_config["name"]
    m = importlib.import_module("pytorch3dunet.unet3d.utils")
    clazz = getattr(m, class_name)
    return clazz(**formatter_config)


def expand_as_one_hot(input, C, ignore_index=None):
    """
    Converts NxSPATIAL label image to NxCxSPATIAL, where each label gets converted to its corresponding one-hot vector.
    It is assumed that the batch dimension is present.
    Args:
        input (torch.Tensor): 3D/4D input image
        C (int): number of channels/labels
        ignore_index (int): ignore index to be kept during the expansion
    Returns:
        4D/5D output torch.Tensor (NxCxSPATIAL)
    """
    assert input.dim() == 4

    # expand the input tensor to Nx1xSPATIAL before scattering
    input = input.unsqueeze(1)
    # create output tensor shape (NxCxSPATIAL)
    shape = list(input.size())
    shape[1] = C

    if ignore_index is not None:
        # create ignore_index mask for the result
        mask = input.expand(shape) == ignore_index
        # clone the src tensor and zero out ignore_index in the input
        input = input.clone()
        input[input == ignore_index] = 0
        # scatter to get the one-hot tensor
        result = torch.zeros(shape).to(input.device).scatter_(1, input, 1)
        # bring back the ignore_index in the result
        result[mask] = ignore_index
        return result
    else:
        # scatter to get the one-hot tensor
        return torch.zeros(shape).to(input.device).scatter_(1, input, 1)


def convert_to_numpy(*inputs):
    """
    Coverts input tensors to numpy ndarrays
    Args:
        inputs (iteable of torch.Tensor): torch tensor
    Returns:
        tuple of ndarrays
    """

    def _to_numpy(i):
        assert isinstance(i, torch.Tensor), "Expected input to be torch.Tensor"
        return i.detach().cpu().numpy()

    return (_to_numpy(i) for i in inputs)


def create_optimizer(optimizer_config, model):
    learning_rate = optimizer_config["learning_rate"]
    weight_decay = optimizer_config.get("weight_decay", 0)
    betas = tuple(optimizer_config.get("betas", (0.9, 0.999)))
    optimizer = optim.Adam(
        model.parameters(),
        lr=learning_rate,
        betas=betas,
        weight_decay=weight_decay,
    )
    return optimizer


def create_lr_scheduler(lr_config, optimizer):
    if lr_config is None:
        return None
    class_name = lr_config.pop("name")
    m = importlib.import_module("torch.optim.lr_scheduler")
    clazz = getattr(m, class_name)
    # add optimizer to the config
    lr_config["optimizer"] = optimizer
    return clazz(**lr_config)


def get_class(class_name, modules):
    for module in modules:
        m = importlib.import_module(module)
        clazz = getattr(m, class_name, None)
        if clazz is not None:
            return clazz
    raise RuntimeError(f"Unsupported dataset class: {class_name}")



import numpy as np
from torch.utils.data import Dataset
from .Tiles_3D import VolumeSlicer


class PredictTiled(Dataset):

    """
    Dataset of fluorescently labeled cell membranes
    """

    def __init__(
        self,
        tiler: VolumeSlicer,
        patch_size=(8, 256, 256),
        overlap=(1, 16, 16),
        crop=(2, 32, 32),
    ):

        # Sanity checks
        assert len(patch_size) == 3, "Patch size must be 3-dimensional."
        # Save parameters
        self.tiler = tiler
        self.patch_size = patch_size
        self.overlap = overlap
        self.crop = crop

    def __len__(self):

        return len(self.tiler.locations)

    def __getitem__(self, idx):

        self.tiler.split(idx)

        return (
            self.tiler.tile[np.newaxis, ...],
            self.tiler.patch_start,
            self.tiler.patch_end,
        )
